--[[
Name: DogTag-1.0
Revision: $Rev: 34589 $
Author: Cameron Kenneth Knight (ckknight@gmail.com)
Inspired By: WatchDog by Vika
Website: http://www.wowace.com/
Documentation: http://www.wowace.com/wiki/DogTag-1.0
SVN: svn://svn.wowace.com/wowace/trunk/DogTag-1.0/DogTag-1.0
Description: A library to provide a markup syntax for unit frame FontStrings
Dependencies: AceLibrary
License: LGPL v2.1
]]

-------------------------------------------------------------------
-- FOR TAG INFORMATION SEE http://www.wowace.com/wiki/DogTag-1.0 --
-------------------------------------------------------------------

local MAJOR_VERSION = "DogTag-1.0"
local MINOR_VERSION = "$Revision: 34589 $"

-- This ensures the code is only executed if the libary doesn't already exist, or is a newer version
if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary.") end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end

local DogTag = {}

local L = {
	-- short races
	["BE"] = "BE",
	["Dr"] = "Dr",
	["Dw"] = "Dw",
	["Gn"] = "Gn",
	["Hu"] = "Hu",
	["NE"] = "NE",
	["Or"] = "Or",
	["Ta"] = "Ta",
	["Tr"] = "Tr",
	["Ud"] = "Ud",
	
	-- short classes
	["Wr"] = "Wr",
	["Pr"] = "Pr",
	["Ma"] = "Ma",
	["Sh"] = "Sh",
	["Pa"] = "Pa",
	["Wl"] = "Wl",
	["Dr"] = "Dr",
	["Ro"] = "Ro",
	["Hu"] = "Hu",
	
	-- classifications
	["Rare"] = ITEM_QUALITY3_DESC,
	["Rare-Elite"] = ITEM_QUALITY3_DESC .. "-" .. ELITE,
	["Elite"] = ELITE,
	["Boss"] = BOSS,
	-- short classifications
	["r"] = "r",
	["r+"] = "r+",
	["+"] = "+",
	["b"] = "b",
	
	["Dead"] = DEAD,
	["Ghost"] = "Ghost",
	["Offline"] = PLAYER_OFFLINE,
	["Combat"] = "Combat",
	["Resting"] = "Resting",
	["Tapped"] = "Tapped",
	
	["Rage"] = RAGE,
	["Focus"] = FOCUS,
	["Energy"] = ENERGY,
	["Mana"] = MANA,
	
	["PvP"] = PVP,
	
	-- genders
	["Male"] = MALE,
	["Female"] = FEMALE,
	
	-- shortgenders
	["m"] = "m",
	["f"] = "f",
	
	["Rested"] = "Rested",
	
	["Leader"] = "Leader",
	["L"] = "L", -- short leader
}
if GetLocale() == "deDE" then
	-- alter L here
end

local RAID_CLASS_COLORS = RAID_CLASS_COLORS
local UNKNOWN = UNKNOWN or "Unknown"
local unpack = unpack

local new, del
do
	local list = setmetatable({}, {__mode='k'})
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return {...}
		end
	end
	function del(t)
		for k in pairs(t) do
			t[k] = nil
		end
		list[t] = true
		return nil
	end
end

local colors = {}
for class, data in pairs(RAID_CLASS_COLORS) do
	colors[class] = { data.r, data.g, data.b, }
end
colors.unknown = { 0.8, 0.8, 0.8 }
colors.hostile = { 226/255, 45/255, 75/255 }
colors.neutral = { 1, 1, 34/255 }
colors.friendly = { 0.2, 0.8, 0.15 }
colors.civilian = { 48/255, 113/255, 191/255 }
colors.tapped = { 0.5, 0.5, 0.5 }
colors.dead = { 0.6, 0.6, 0.6 }
colors.disconnected = { 0.7, 0.7, 0.7 }
colors.inCombat = { 1, 0, 0 }
colors.resting = { 1, 1, 0 }
colors.petHappy = { 0, 1, 0 }
colors.petNeutral = { 1, 1, 0 }
colors.petAngry = { 1, 0, 0 }
colors.rage = { 226/255, 45/255, 75/255 }
colors.energy = { 1, 220/255, 25/255 }
colors.focus = { 1, 210/255, 0 }
colors.mana = { 48/255, 113/255, 191/255 }
colors.minHP = { 1, 0, 0 }
colors.midHP = { 1, 1, 0 }
colors.maxHP = { 0, 1, 0 }

local UnitCanAttack = UnitCanAttack
local UnitIsPVP = UnitIsPVP
local UnitPlayerControlled = UnitPlayerControlled
local UnitIsPlayer = UnitIsPlayer
local UnitClass = UnitClass
local UnitHealth = UnitHealth
local UnitHealthMax = UnitHealthMax
local UnitMana = UnitMana
local UnitManaMax = UnitManaMax
local UnitAffectingCombat = UnitAffectingCombat
local UnitIsDeadOrGhost = UnitIsDeadOrGhost
local UnitIsDead = UnitIsDead
local UnitIsGhost = UnitIsGhost
local UnitIsConnected = UnitIsConnected
local IsResting = IsResting
local GetPetHappiness = GetPetHappiness
local UnitIsTapped = UnitIsTapped
local UnitIsTappedByPlayer = UnitIsTappedByPlayer
local UnitReaction = UnitReaction
local UnitLevel = UnitLevel
local GetDifficultyColor = GetDifficultyColor
local UnitPowerType = UnitPowerType
local UnitIsFriend = UnitIsFriend
local UnitCreatureFamily = UnitCreatureFamily
local UnitCreatureType = UnitCreatureType
local GetComboPoints = GetComboPoints
local UnitClassification = UnitClassification
local UnitFactionGroup = UnitFactionGroup
local UnitRace = UnitRace
local UnitIsPlusMob = UnitIsPlusMob
local UnitSex = UnitSex
local GetRestState = GetRestState
local UnitIsPartyLeader = UnitIsPartyLeader
local GetPetExperience = GetPetExperience
local UnitXP = UnitXP
local UnitXPMax = UnitXPMax
local GetXPExhaustion = GetXPExhaustion
local GetNumRaidMembers = GetNumRaidMembers
local GetRaidRosterInfo = GetRaidRosterInfo
local UnitName = UnitName
local GetGuildInfo = GetGuildInfo

local function GetClassColor(unit)
	local _,class = UnitClass(unit)
	return unpack(colors[class] or colors.unknown)
end

local function GetHostilityColor(unit)
	if UnitIsPlayer(unit) or UnitPlayerControlled(unit) then
		if UnitCanAttack(unit, "player") then
			-- they can attack me
			if UnitCanAttack("player", unit) then
				-- and I can attack them
				return unpack(colors.hostile)
			else
				-- but I can't attack them
				return unpack(colors.civilian)
			end
		elseif UnitCanAttack("player", unit) then
			-- they can't attack me, but I can attack them
			return unpack(colors.neutral)
		elseif UnitIsPVP(unit) then
			-- on my team
			return unpack(colors.friendly)
		else
			-- either enemy or friend, no violence
			return unpack(colors.civilian)
		end
	elseif (UnitIsTapped(unit) and not UnitIsTappedByPlayer(unit)) or UnitIsDead(unit) then
		return unpack(colors.tapped)
	else
		local reaction = UnitReaction(unit, "player")
		if reaction then
			if reaction >= 5 then
				return unpack(colors.friendly)
			elseif reaction == 4 then
				return unpack(colors.neutral)
			else
				return unpack(colors.hostile)
			end
		else
			return unpack(colors.unknown)
		end
	end
end

local function HealthGradient(perc)
	local r1, g1, b1
	local r2, g2, b2
	if perc <= 0.5 then
		perc = perc * 2
		r1, g1, b1 = unpack(colors.minHP)
		r2, g2, b2 = unpack(colors.midHP)
	else
		perc = perc * 2 - 1
		r1, g1, b1 = unpack(colors.midHP)
		r2, g2, b2 = unpack(colors.maxHP)
	end
	return r1 + (r2-r1)*perc, g1 + (g2-g1)*perc, b1 + (b2-b1)*perc
end
-- Note, these tables use the unlocalized second returns of UnitClass and UnitRace, thus the keys are not localized.
local shortraces = {
	BloodElf = L["BE"],
	Draenei = L["Dr"],
	Dwarf = L["Dw"],
	Gnome = L["Gn"],
	Human = L["Hu"],
	NightElf = L["NE"],
	Orc = L["Or"],
	Tauren = L["Ta"],
	Troll = L["Tr"],
	Undead = L["Ud"],
}
local shortclasses = {
	WARRIOR = L["Wr"],
	PRIEST = L["Pr"],
	MAGE = L["Ma"],
	SHAMAN = L["Sh"],
	PALADIN = L["Pa"],
	WARLOCK = L["Wl"],
	DRUID = L["Dr"],
	ROGUE = L["Ro"],
	HUNTER = L["Hu"],
}

local lastMouseFocus

local Modifiers; Modifiers = {}
-- when writing modifiers, inputs can be a number or a string. Also, any of these can be returned, as well as nil, which means a blank string.
-- the modifier itself can be in the form of a function or a table containing a function and an event string (or a table of event strings). The events specified will cause the string to update upon that event.
-- note: can take an argument, either nil, a number, a string.
Modifiers.isequal = {"if value ~= ${arg} then value = nil end", value = "number;string", arg = "number;string", ret = "same;nil"}

Modifiers.isless = {"if type(value) ~= ${argtype} then if value >= ${arg} then value = nil end else if tostring(value) >= ${arg:string} then value = nil end end", value = "number;string", arg = "number;string", ret = "same;nil", globals = "tostring"}

Modifiers.islessequal = {"if type(value) ~= ${argtype} then if value > ${arg} then value = nil end else if tostring(value) > ${arg:string} then value = nil end end", value = "number;string", arg = "number;string", ret = "same;nil", globals = "tostring"}

Modifiers.isgreater = "~islessequal"
Modifiers.isgreaterequal = "~isless"
Modifiers.isnotequal = "~isequal"

Modifiers.hide = "~isequal"
Modifiers.hideless = "~isless"
Modifiers.hidegreater = "~isgreater"
Modifiers.hidelessequal = "~islessequal"
Modifiers.hidegreaterequal = "~isgreaterequal"
Modifiers.hideunless = "~isnotequal"
Modifiers.hidezero = "hide(0)"

Modifiers.prepend = {"value = ${arg} .. value", value = "number;string", arg = "string", ret = "string"}

Modifiers.append = {"value = value .. ${arg}", value = "number;string", arg = "string", ret = "string"}

Modifiers.negate = {"value = -value", value = "number", ret = "number"}

Modifiers.percent = {"value = value .. '%'", value = "number", ret = "string"}

Modifiers.short = {[[
	if type(value) ~= "number" then
		local a,b = value:match("^(%d+)/(%d+)")
		if a then
			a, b = tonumber(a), tonumber(b)
			if b >= 10000000 then
				b = ("%.1fm"):format(b / 1000000)
			elseif b >= 1000000 then
				b = ("%.2fm"):format(b / 1000000)
			elseif b >= 100000 then
				b = ("%.0fk"):format(b / 1000)
			elseif b >= 10000 then
				b = ("%.1fk"):format(b / 1000)
			end
			if a >= 10000000 then
				a = ("%.1fm"):format(a / 1000000)
			elseif a >= 1000000 then
				a = ("%.2fm"):format(a / 1000000)
			elseif a >= 100000 then
				a = ("%.0fk"):format(a / 1000)
			elseif a >= 10000 then
				a = ("%.1fk"):format(a / 1000)
			end
			value = a.."/"..b
		end
	else	
		if value >= 10000000 then
			value = ("%.1fm"):format(value / 1000000)
		elseif value >= 1000000 then
			value = ("%.2fm"):format(value / 1000000)
		elseif value >= 100000 then
			value = ("%.0fk"):format(value / 1000)
		elseif value >= 10000 then
			value = ("%.1fk"):format(value / 1000)
		end
	end
]],	value = "number;string", ret = "same;nil"}

Modifiers.veryshort = {[[
	if type(value) ~= "number" then
		local a,b = value:match("^(%d+)/(%d+)")
		if a then
			a, b = tonumber(a), tonumber(b)
			if b >= 1000000 then
				b = ("%.0fm"):format(b / 1000000)
			elseif b >= 1000 then
				b = ("%.0fk"):format(b / 1000)
			end
			if a >= 1000000 then
				a = ("%.0fm"):format(a / 1000000)
			elseif a >= 1000 then
				a = ("%.0fk"):format(a / 1000)
			end
			value = a.."/"..b
		end
	else	
		if value >= 1000000 then
			value = ("%.0fm"):format(value / 1000000)
		elseif value >= 1000 then
			value = ("%.0fk"):format(value / 1000)
		end
	end
]], value = "number;string", ret = "same;nil"}

Modifiers.upper = {"value = value:upper()", value = "string", ret = "string"}

Modifiers.lower = {"value = value:lower()", value = "string", ret = "string"}

Modifiers.bracket = {"value = '[' .. value .. ']'", value = "number;string", ret = "string"}

Modifiers.paren = {"value = '(' .. value .. ')'", value = "number;string", ret = "string"}

Modifiers.angle = {"value = '<' .. value .. '>'", value = "number;string", ret = "string"}

Modifiers.brace = {"value = '{' .. value .. '}'", value = "number;string", ret = "string"}

Modifiers.trunc = {[[do
	local len = 0
	for i = 1, ${arg} do
		local b = value:byte(len+1)
		if not b then
			break
		elseif b <= 127 then
			len = len + 1
		elseif b <= 223 then
			len = len + 2
		elseif b <= 239 then
			len = len + 3
		else
			len = len + 4
		end
	end
	value = value:sub(1, len)
end]], arg = "number", value = "string", ret = "string"}

Modifiers.contains = {"if not value:find(${arg}) then value = nil end", value = "string", arg = "string", ret = "string;nil"}

Modifiers.hidecontains = "~contains"

Modifiers.repcolor = {function(data)
	return [[do
		local repname, repreaction, repmin, repmax, repvalue = GetWatchedFactionInfo()
		if not repname then
		]] .. (data.isMod and [[
			value = "|r" .. value
		]] or [[
			value = "|r"
		]]) .. [[
		else
			local info = FACTION_BAR_COLORS[repreaction]
		]] .. (data.isMod and [[
			value = ("|cff%02x%02x%02x%s|r"):format(info.r * 255, info.g * 255, info.b * 255, value)
		]] or [[
			value = ("|cff%02x%02x%02x"):format(info.r * 255, info.g * 255, info.b * 255)
		]]) .. [[
		end
	end]]
end, events = "Reputation", globals = "GetWatchedFactionInfo;FACTION_BAR_COLORS"}

Modifiers.statuscolor = {function(data)
	return [[do
		local r,g,b
		if not UnitIsConnected(${unit}) then
			r,g,b = unpack(colors.disconnected)
		elseif UnitIsDeadOrGhost(${unit}) then
			r,g,b = unpack(colors.dead)
		elseif UnitAffectingCombat(${unit}) then
			r,g,b = unpack(colors.inCombat)
		elseif ${unit} == "player" and IsResting() then
			r,g,b = unpack(colors.resting)
		else
		]] .. (data.isMod and [[
			value = "|r" .. value
		]] or [[
			value = "|r"
		]]) .. [[
		end
		if r then
		]] .. (data.isMod and [[
			value = ("|cff%02x%02x%02x%s|r"):format(r * 255, g * 255, b * 255, value)
		]] or [[
			value = ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
		]]) .. [[
		end
	end]]
end, value = "number;string", ret = "string", events = "Update" --[[due to the combat thing]], globals = "unpack;UnitIsConnected;UnitIsDeadOrGhost;UnitAffectingCombat;IsResting"}

Modifiers.happycolor = {function(data)
	return [[do
		local x = GetPetHappiness()
		local r,g,b
		if x == 3 then
			r,g,b = unpack(colors.petHappy)
		elseif x == 2 then
			r,g,b = unpack(colors.petNeutral)
		elseif x == 1 then
			r,g,b = unpack(colors.petAngry)
		else
		]] .. (data.isMod and [[
			value = "|r" .. value
		]] or [[
			value = "|r"
		]]) .. [[
		end
		if r then
		]] .. (data.isMod and [[
			value = ("|cff%02x%02x%02x%s|r"):format(r * 255, g * 255, b * 255, value)
		]] or [[
			value = ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
		]]) .. [[
		end
	end]]
end, value = "number;string", ret = "string", events = "UNIT_HAPPINESS", globals = "unpack;GetPetHappiness" }

Modifiers.hostilecolor = {function(data)
	return [[do
		local r, g, b = DogTag___GetHostilityColor(${unit})
	]] .. (data.isMod and [[
		value = ("|cff%02x%02x%02x%s|r"):format(r * 255, g * 255, b * 255, value)
	]] or [[
		value = ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
	]]) .. [[
	end]]
end, value = "number;string", ret = "string", events = "UNIT_FACTION", globals = "DogTag.__GetHostilityColor" }

Modifiers.classcolor = {function(data)
	return [[do
		local r, g, b = DogTag___GetClassColor(${unit})
	]] .. (data.isMod and [[
		value = ("|cff%02x%02x%02x%s|r"):format(r * 255, g * 255, b * 255, value)
	]] or [[
		value = ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
	]]) .. [[
	end]]
end, value = "number;string", ret = "string", globals = "DogTag.__GetClassColor"}

Modifiers.hpcolor = {function(data)
	return [[do
		local perc = UnitHealth(${unit}) / UnitHealthMax(${unit})
		local r, g, b = DogTag___HealthGradient(perc)
	]] .. (data.isMod and [[
		value = ("|cff%02x%02x%02x%s|r"):format(r * 255, g * 255, b * 255, value)
	]] or [[
		value = ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
	]]) .. [[
	end]]
end, value = "number;string", ret = "string", events = "UNIT_HEALTH;UNIT_MAXHEALTH", globals = "UnitHealth;UnitHealthMax;DogTag.__HealthGradient" }

Modifiers.aggrocolor = {function(data)
	return [[do
		local val = UnitReaction("player", ${unit}) or 5

		local info = UnitReactionColor[val]
	]] .. (data.isMod and [[
		value = ("|cff%02x%02x%02x%s|r"):format(info.r * 255, info.g * 255, info.b * 255, value)
	]] or [[
		value = ("|cff%02x%02x%02x"):format(info.r * 255, info.g * 255, info.b * 255)
	]]) .. [[
	end]]
end, value = "number;string", ret = "string", events = "UNIT_FACTION", globals = "UnitReaction;UnitReactionColor" }

Modifiers.difficultycolor = {function(data)
	return [[if UnitCanAttack("player", ${unit}) then
		local level = UnitLevel(${unit})
		if level == 0 then
			level = 99
		end
		local info = GetDifficultyColor(level)
	]] .. (data.isMod and [[
		value = ("|cff%02x%02x%02x%s|r"):format(info.r * 255, info.g * 255, info.b * 255, value)
	]] or [[
		value = ("|cff%02x%02x%02x"):format(info.r * 255, info.g * 255, info.b * 255)
	]]) .. [[
	end]]
end, value = "number;string", ret = "string", events = "UNIT_LEVEL;PLAYER_LEVEL_UP", globals = "UnitCanAttack;UnitLevel;GetDifficultyColor" }

Modifiers.manacolor = {function(data)
	return [[do
		local powerType = UnitPowerType(${unit})
		local r,g,b
		if powerType == 0 then
			r,g,b = unpack(colors.mana)
		elseif powerType == 1 then
			r,g,b = unpack(colors.rage)
		elseif powerType == 2 then
			r,g,b = unpack(colors.focus)
		elseif powerType == 3 then
			r,g,b = unpack(colors.energy)
		else
			r,g,b = unpack(colors.unknown)
		end
	]] .. (data.isMod and [[
		value = ("|cff%02x%02x%02x%s|r"):format(r * 255, g * 255, b * 255, value)
	]] or [[
		value = ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
	]]) .. [[
	end]]
end, value = "number;string", ret = "string", events = "UNIT_DISPLAYPOWER", globals = "unpack;UnitPowerType" }

Modifiers.combatcolor = {function(data)
	return [[
		if UnitAffectingCombat(${unit}) then
			local r,g,b = unpack(colors.inCombat)
	]] .. (data.isMod and [[
			value = ("|cff%02x%02x%02x%s|r"):format(r * 255, g * 255, b * 255, value)
	]] or [[
			value = ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
	]]) .. [[
		else	
	]] .. (data.isMod and [[
			value = "|r" .. value
		]] or [[
			value = "|r"
	]]) .. [[
		end
	]]
end, value = "number;string", ret = "string", events = "Update", globals = "unpack;UnitAffectingCombat" }

Modifiers.round = {function(data)
	local arg = math.floor((data.arg or 0) + 0.5)
	if arg == 0 then
		return [[value = math_floor(value + 0.5)]]
	else
		return [[value = math_floor(value * (10^]] .. arg .. [[) + 0.5) / (10^]] .. arg .. [[)]]
	end
end, arg = "nil;number", value = "number", ret = "number", globals = "math.floor"}

Modifiers.color = {function(data)
	local arg = data.arg:lower()
	if not arg:match("^%x%x%x%x%x%x$") then
		arg = "ffffff"
	end
	if data.isMod then
		return [[value = "|cff]] .. arg .. [[" .. value .. "|r"]]
	else
		return [[value = "|cff]] .. arg .. [["]]
	end
end, arg = "string", value = "number;string", ret = "string"}
-- colors
Modifiers.white = "color(ffffff)"
Modifiers.red = "color(ff0000)"
Modifiers.green = "color(00ff00)"
Modifiers.blue = "color(0000ff)"
Modifiers.cyan = "color(00ffff)"
Modifiers.fuchsia = "color(ff00ff)"
Modifiers.yellow = "color(ffff00)"
Modifiers.gray = "color(afafaf)"

local Tags = {}
-- When writing tags, `unit` is modified before the call, that contains the unit id.
-- Also, nil (which translates to empty string), a number, or a string can be returned.
-- the tag itself can be in the form of a function or a table containing a function and an event string (or a table of event strings). The events specified will cause the string to update upon that event.
-- Note: can take an argument, either nil, a number, or a string.

Tags.isfriend = {"value = UnitIsFriend('player', ${unit}) and 'true'", ret = "string;nil", events = "UNIT_FACTION", globals = "UnitIsFriend" }

Tags.isenemy = "~isfriend"

Tags.text = {"value = ${arg}", arg = "string", ret = "string"}

Tags.name = {"value = UnitName(${unit})", ret = "string", events = "UNIT_NAME_UPDATE", globals = "UnitName"}

Tags.namehostile = "name:hostilecolor"

Tags.status = {[[do
	if not UnitIsConnected(${unit}) then
		value = L["Offline"]
	elseif UnitIsDead(${unit}) then
		value = L["Dead"]
	elseif UnitIsGhost(${unit}) then
		value = L["Ghost"]
	elseif UnitAffectingCombat(${unit}) then
		value = L["Combat"]
	elseif ${unit} == "player" and IsResting() then
		value = L["Resting"]
	end
end]], ret = "string;nil", events = "Update" --[[due to the combat thing]], globals = "UnitIsConnected;UnitIsDead;UnitIsGhost;UnitAffectingCombat;IsResting"}

Tags.statusgone = {[[do
	if not UnitIsConnected(${unit}) then
		value = L["Offline"]
	elseif UnitIsDead(${unit}) then
		value = L["Dead"]
	elseif UnitIsGhost(${unit}) then
		value = L["Ghost"]
	end
end]], ret = "string;nil", events = "PLAYER_ALIVE;PLAYER_DEAD;PLAYER_UNGHOST", globals = "UnitIsConnected;UnitIsDead;UnitIsGhost" }

Tags.curhp = {function(data)
	if MobHealth3 then
		return [[do
			local currValue = UnitHealth(${unit})
			if currValue == 0 then
				value = 0
			else
				if not UnitIsFriend("player", ${unit}) then
					local maxValue = UnitHealthMax(${unit})
					local curr, max, MHfound = MobHealth3:GetUnitHealth(${unit}, currValue, maxValue)
					if MHfound then
						value = curr
					end
				end
				if not value then
					value = currValue
				end
			end
		end]]
	elseif MobHealth_PPP then
		return [[do
			local currValue = UnitHealth(${unit})
			if currValue == 0 then
				value = 0
			else
				if not UnitIsFriend("player", ${unit}) then
					local name = UnitName(${unit})
					local level = UnitLevel(${unit})
					local ppp = MobHealth_PPP(name..":"..level)
					if ppp > 0 then
						value = math_floor(currValue * ppp + 0.5)
					end
				end
				if not value then
					value = currValue
				end
			end
		end]]
	else
		return [[value = UnitHealth(${unit})]]
	end
end, ret = "number", events = "UNIT_HEALTH;UNIT_MAXHEALTH", globals = "UnitHealth;UnitIsFriend;UnitHealthMax;MobHealth3;MobHealth_PPP;UnitName;UnitLevel;math.floor" }

Tags.maxhp = {function(data)
	if MobHealth3 then
		return [[do
			local maxValue = UnitHealthMax(${unit})
			if maxValue == 0 then
				value = 0
			else
				if not UnitIsFriend("player", ${unit}) then
					local curr, max, MHfound = MobHealth3:GetUnitHealth(${unit}, 1, maxValue)
					if MHfound then
						value = max
					end
				end
				if not value then
					value = maxValue
				end
			end
		end]]
	elseif MobHealth_PPP then
		return [[do
			if not UnitIsFriend("player", ${unit}) then
				local name = UnitName(${unit})
				local level = UnitLevel(${unit})
				local ppp = MobHealth_PPP(name..":"..level)
				if ppp > 0 then
					value = math_floor(100 * ppp + 0.5)
				end
			end
			if not value then
				value = UnitHealthMax(${unit})
			end
		end]]
	else
		return [[value = UnitHealthMax(${unit})]]
	end
end, ret = "number", events = "UNIT_HEALTH;UNIT_MAXHEALTH", globals = "UnitHealth;UnitIsFriend;UnitHealthMax;MobHealth3;MobHealth_PPP;UnitName;UnitLevel;math.floor" }

Tags.percenthp = {[[do
	local hpmax = UnitHealthMax(${unit})
	if hpmax == 0 then
		value = 0
	else
		value = math_floor(1000 * UnitHealth(${unit}) / hpmax + 0.5) / 10
	end
end]], ret = "number", events = "UNIT_HEALTH;UNIT_MAXHEALTH", globals = "UnitHealth;UnitHealthMax;math.floor" }

Tags.missinghp = {function(data)
	if MobHealth3 then
		return [[do
			local maxValue = UnitHealthMax(${unit})
			if maxValue == 0 then
				value = 0
			else
				local currValue = UnitHealth(${unit})
				if MobHealth3 and not UnitIsFriend("player", ${unit}) then
					local curr, max, MHfound = MobHealth3:GetUnitHealth(${unit}, currValue, maxValue)
					if MHfound then
						value = max - curr
					end
				end
				if not value then
					value = maxValue - currValue
				end
			end
		end]]
	elseif MobHealth_PPP then
		return [[do
			local currValue = UnitHealth(${unit})
			if currValue == 0 then
				value = 0
			else
				if not UnitIsFriend("player", ${unit}) then
					local name = UnitName(${unit})
					local level = UnitLevel(${unit})
					local ppp = MobHealth_PPP(name..":"..level)
					if ppp > 0 then
						value = math_floor((100 - currValue) * ppp + 0.5)
					end
				end
				if not value then
					value = UnitHealthMax(${unit}) - currValue
				end
			end
		end]]
	else
		return [[value = UnitHealthMax(${unit}) - UnitHealth(${unit})]]
	end
end, ret = "number", events = "UNIT_HEALTH;UNIT_MAXHEALTH", globals = "UnitHealth;UnitHealthMax;UnitIsFriend;MobHealth3;MobHealth_PPP;UnitName;UnitLevel;math.floor" }

Tags.fractionalhp = {function(data)
	if MobHealth3 then
		return [[do
			local maxValue = UnitHealthMax(${unit})
			if maxValue == 0 then
				value = "0/0"
			else
				local currValue = UnitHealth(${unit})
				if maxValue > 100 then
					value = currValue.."/"..maxValue
				elseif MobHealth3 and not UnitIsFriend("player", ${unit}) then
					local curr, max, MHfound = MobHealth3:GetUnitHealth(${unit}, currValue, maxValue)
					if MHfound then
						value = curr.."/"..max
					end
				end
				if not value then
					value = currValue.."/"..maxValue
				end
			end
		end]]
	elseif MobHealth_PPP then
		return [[do
			local currValue = UnitHealth(${unit})
			if not UnitIsFriend("player", ${unit}) then
				local name = UnitName(${unit})
				local level = UnitLevel(${unit})
				local ppp = MobHealth_PPP(name..":"..level)
				if ppp > 0 then
					value = math_floor(currValue * ppp + 0.5) .. "/" .. math_floor(100 * ppp + 0.5)
				end
			end
			if not value then
				value = currValue .. "/" .. UnitHealthMax(${unit})
			end
		end]]
	else
		return [[value = UnitHealth(${unit}).."/"..UnitHealthMax(${unit})]]
	end
end, ret = "string", events = "UNIT_HEALTH;UNIT_MAXHEALTH", globals = "UnitHealth;UnitHealthMax;UnitIsFriend;MobHealth3;MobHealth_PPP;UnitName;UnitLevel;math.floor" }

Tags.smarthp = {function(data)
	if MobHealth3 then
		return [[do
			local maxValue = UnitHealthMax(${unit})
			if maxValue == 0 then
				value = "0%"
			else
				local currValue = UnitHealth(${unit})
				if maxValue > 100 then
					value = currValue.."/"..maxValue
				end
				if MobHealth3 and not UnitIsFriend("player", ${unit}) then
					local curr, max, MHfound = MobHealth3:GetUnitHealth(${unit}, currValue, maxValue)
					if MHfound then
						value = curr.."/"..max
					end
				end
				if not value then
					value = ("%.1f%%"):format(currValue / maxValue * 100)
				end
			end
		end]]
	elseif MobHealth_PPP then
		return [[do
			if not UnitIsFriend("player", ${unit}) then
				local name = UnitName(${unit})
				local level = UnitLevel(${unit})
				local ppp = MobHealth_PPP(name..":"..level)
				if ppp > 0 then
					value = math_floor(UnitHealth(${unit}) * ppp + 0.5) .. "/" .. math_floor(100 * ppp + 0.5)
				end
			end
			if not value then
				value = ("%.1f%%"):format(UnitHealth(${unit}) / UnitHealthMax(${unit}) * 100)
			end
		end]]
	else
		return [[do
			local maxValue = UnitHealthMax(${unit})
			if maxValue == 0 then
				value = "0%"
			else
				local currValue = UnitHealth(${unit})
				if maxValue > 100 then
					value = currValue.."/"..maxValue
				else
					value = ("%.1f%%"):format(currValue / maxValue * 100)
				end
			end
		end]]
	end
end, ret = "string", events = "UNIT_HEALTH;UNIT_MAXHEALTH", globals = "UnitHealth;UnitHealthMax;UnitIsFriend;MobHealth3;MobHealth_PPP;UnitName;UnitLevel;math.floor" }

Tags.surehp = {function(data)
	if MobHealth3 then
		return [[do
			local maxValue = UnitHealthMax(${unit})
			if maxValue ~= 0 then
				local currValue = UnitHealth(${unit})
				if maxValue > 100 then
					value = currValue.."/"..maxValue
				end
				if MobHealth3 and not UnitIsFriend("player", ${unit}) then
					local curr, max, MHfound = MobHealth3:GetUnitHealth(${unit}, currValue, maxValue)
					if MHfound then
						value = curr.."/"..max
					end
				end
			end
		end]]
	elseif MobHealth_PPP then
		return [[do
			if not UnitIsFriend("player", ${unit}) then
				local name = UnitName(${unit})
				local level = UnitLevel(${unit})
				local ppp = MobHealth_PPP(name..":"..level)
				if ppp > 0 then
					value = math_floor(UnitHealth(${unit}) * ppp + 0.5) .. "/" .. math_floor(100 * ppp + 0.5)
				end
			else
				local maxValue = UnitHealthMax(${unit})
				if maxValue > 100 then
					value = UnitHealth(${unit}).."/"..maxValue
				end
			end
		end]]
	else
		return [[do
			local maxValue = UnitHealthMax(${unit})
			if maxValue ~= 0 then
				local currValue = UnitHealth(${unit})
				if maxValue > 100 then
					value = currValue.."/"..maxValue
				end
			end
		end]]
	end
end, ret = "string;nil", events = "UNIT_HEALTH;UNIT_MAXHEALTH", globals = "UnitHealth;UnitHealthMax;UnitIsFriend;MobHealth3;MobHealth_PPP;UnitName;UnitLevel;math.floor" }

Tags.curmana = {"value = UnitMana(${unit})", ret = "number", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER", globals = "UnitMana" }

Tags.maxmana = {"value = UnitManaMax(${unit})", ret = "number", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER", globals = "UnitManaMax" }

Tags.percentmana = {[[do
	local mpmax = UnitManaMax(${unit})
	if mpmax == 0 then
		value = 0
	end
	value = math_floor(1000 * UnitMana(${unit}) / mpmax + 0.5) / 10
end]], ret = "number", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER", globals = "UnitMana;UnitManaMax;math.floor" }

Tags.missingmana = {"value = UnitManaMax(${unit}) - UnitMana(${unit})", ret = "number", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER", globals = "UnitMana;UnitManaMax" }

Tags.fractionalmana = {"value = UnitMana(${unit})..'/'..UnitManaMax(${unit})", ret = "string", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER", globals = "UnitMana;UnitManaMax" }

Tags.typepower = {[[do
	local p = UnitPowerType(${unit})
	if p == 1 then
		value = L["Rage"]
	elseif p == 2 then
		value = L["Focus"]
	elseif p == 3 then
		value = L["Energy"]
	else
		value = L["Mana"]
	end
end]], ret = "string", events = "UNIT_DISPLAYPOWER", globals = "UnitPowerType" }
Tags.typemana = "typepower"

Tags.ispowertype = {function(data)
	local arg = data.arg:lower()
	if arg == "rage" then
		return [[value = UnitPowerType(${unit}) == 1 and 'rage']]
	elseif arg == "focus" then
		return [[value = UnitPowerType(${unit}) == 2 and 'focus']]
	elseif arg == "energy" then
		return [[value = UnitPowerType(${unit}) == 3 and 'energy']]
	elseif arg == "mana" then
		return [[value = UnitPowerType(${unit}) == 0 and 'mana']]
	else
		return [[value = nil]]
	end
end, arg = "string", ret = "nil;string", events = "UNIT_DISPLAYPOWER", globals = "UnitPowerType"}

Tags.israge = "ispowertype(rage)"
Tags.isfocus = "ispowertype(focus)"
Tags.isenergy = "ispowertype(energy)"
Tags.ismana = "ispowertype(mana)"

Tags.isplayerpowertype = "ispowertype#player"

Tags.isplayerrage = "israge#player"
Tags.isplayerfocus = "isfocus#player"
Tags.isplayerenergy = "isenergy#player"
Tags.isplayermana = "ismana#player"

Tags.level = {[[do
	local x = UnitLevel(${unit})
	if x > 0 then
		value = x
	else
		value = "??"
	end
end]], ret = "number;string", events = "UNIT_LEVEL", globals = "UnitLevel" }

Tags.class = {[[value = UnitClass(${unit}) or UNKNOWN]], ret = "string", globals = "UnitClass;UNKNOWN"}

Tags.smartclass = {[[if UnitIsPlayer(${unit}) or (not UnitIsFriend("player", ${unit}) and not UnitPlayerControlled(${unit})) then
	value = UnitClass(${unit})
end]], ret = "string;nil", globals = "UnitIsPlayer;UnitIsFriend;UnitPlayerControlled;UnitClass"}

Tags.playerclass = {[[if UnitIsPlayer(${unit}) then
	value = UnitClass(${unit})
end]], ret = "string;nil", globals = "UnitIsPlayer;UnitClass"}

Tags.shortclass = {[[do
	local _,c = UnitClass(${unit})
	if c then
		value = DogTag___shortclasses[c]
	end
end]], ret = "string;nil", globals = "UnitClass;DogTag.__shortclasses"}

Tags.shortsmartclass = {[[do
	if UnitIsPlayer(${unit}) or (not UnitIsFriend("player", ${unit}) and not UnitPlayerControlled(${unit})) then
		local _,c = UnitClass(${unit})
		if c then
			value = DogTag___shortclasses[c]
		end
	end
end]], ret = "string;nil", globals = "UnitClass;UnitIsPlayer;UnitIsFriend;UnitPlayerControlled;DogTag.__shortclasses"}

Tags.shortplayerclass = {[[do
	if UnitIsPlayer(${unit}) then
		local _,c = UnitClass(${unit})
		if c then
			value = DogTag___shortclasses[c]
		end
	end
end]], ret = "string;nil", globals = "UnitIsPlayer;UnitClass;DogTag.__shortclasses"}

Tags.creature = {[[value = UnitCreatureFamily(${unit}) or UnitCreatureType(${unit}) or UNKNOWN]], ret = "string", globals = "UnitCreatureFamily;UnitCreatureType"}

Tags.combos = {[[value = GetComboPoints()]], ret = "number", events = "PLAYER_COMBO_POINTS", globals = "GetComboPoints" }

Tags.combosymbols = {[[value = ("@"):rep(GetComboPoints())]], ret = "string;nil", events = "PLAYER_COMBO_POINTS", globals = "GetComboPoints" }

Tags.classification = {[[do
	local c = UnitClassification(${unit})
	if c == "rare" then
		value = L["Rare"]
	elseif c == "rareelite" then
		value = L["Rare-Elite"]
	elseif c == "elite" then
		value = L["Elite"]
	elseif c == "worldboss" then
		value = L["Boss"]
	end
end]], ret = "string;nil", globals = "UnitClassification"}

Tags.shortclassification = {[[do
	local c = UnitClassification(${unit})
	if c == "rare" then
		value = L["r"]
	elseif c == "rareelite" then
		value = L["r+"]
	elseif c == "elite" then
		value = L["+"]
	elseif c == "worldboss" then
		value = L["b"]
	end
end]], ret = "string;nil", globals = "UnitClassification"}

Tags.faction = {[[do
	local _, faction = UnitFactionGroup(${unit})
	value = faction
end]], ret = "string", globals = "UnitFactionGroup"}

Tags.connect = {[[
	if not UnitIsConnected(${unit}) then
		value = L["Offline"]
	end
]], ret = "string;nil", globals = "UnitIsConnected"}

Tags.smartrace = {[[
	if UnitIsPlayer(${unit}) then
		value = UnitRace(${unit}) or UNKNOWN
	else
		value = UnitCreatureFamily(${unit}) or UnitCreatureType(${unit}) or UNKNOWN
	end
]], ret = "string", globals = "UnitIsPlayer;UnitRace;UnitCreatureFamily;UnitCreatureType;UNKNOWN"}

Tags.race = {"value = UnitRace(${unit})", ret = "string", globals = "UnitRace"}

Tags.shortrace = {[[do
	local _,r = UnitRace(${unit})
	value = DogTag___shortraces[r]
end]], ret = "string;nil", globals = "UnitRace;DogTag.__shortraces" }

Tags.pvp = {[=[value = UnitIsPVP(${unit}) and L["PvP"]]=], ret = "string;nil", globals = "UnitIsPVP"}

Tags.plus = {[[value = UnitIsPlusMob(${unit}) and "+"]], ret = "string;nil", globals = "UnitIsPlusMob"}

Tags.sex = {[[do
	local sex = UnitSex(${unit})
	if sex == 0 then
		value = L["Male"]
	elseif sex == 1 then
		value = L["Female"]
	end
end]], ret = "string;nil", globals = "UnitSex"}

Tags.shortsex = {[[do
	local sex = UnitSex(${unit})
	if sex == 0 then
		value = L["m"]
	elseif sex == 1 then
		value = L["f"]
	end
end]], ret = "string;nil", globals = "UnitSex"}

Tags.rested = {[=[value = GetRestState() and L["Rested"]]=], ret = "string;nil", events = "PLAYER_XP_UPDATE", globals = "GetRestState" }

Tags.leader = {[=[value = UnitIsPartyLeader(${unit}) and L["L"]]=], ret = "string;nil", globals = "UnitIsPartyLeader"}

Tags.leaderlong = {[=[value = UnitIsPartyLeader(${unit}) and L["Leader"]]=], ret = "string;nil", globals = "UnitIsPartyLeader"}

Tags.happynum = {[[value = GetPetHappiness() or 0]], ret = "number", events = "UNIT_HAPPINESS", globals = "GetPetHappiness"}

Tags.happytext = {[=[value = _G["PET_HAPPINESS" .. (GetPetHappiness() or 0)]]=], ret = "string;nil", events = "UNIT_HAPPINESS", globals = "_G;GetPetHappiness" }

Tags.happyicon = {[[do
	local happy = GetPetHappiness()
	if happy == 3 then
		value = ":D"
	elseif happy == 2 then
		value = ":||"
	elseif happy == 1 then
		value = "B("
	end
end]], ret = "string;nil", events = "UNIT_HAPPINESS", globals = "GetPetHappiness" }

Tags.curxp = {[[
	if ${unit} == "pet" then
		value = GetPetExperience()
	else
	 	value = UnitXP(${unit})
	end
]], ret = "number", events = "UNIT_PET_EXPERIENCE;PLAYER_XP_UPDATE", globals = "GetPetExperience;UnitXP" }

Tags.maxxp = {[[
	if ${unit} == "pet" then
		local _,max = GetPetExperience()
		value = max
	else
	 	value = UnitXPMax(${unit})
	end
]], ret = "number", events = "UNIT_PET_EXPERIENCE;PLAYER_XP_UPDATE", globals = "GetPetExperience;UnitXP" }

Tags.fractionalxp = {[[
	local cur, max
	if ${unit} == "pet" then
		cur, max = GetPetExperience()
	else
	 	cur, max = UnitXP(${unit}), UnitXPMax(${unit})
	end	
	value = cur .. '/' .. max
]], ret = "string", events = "UNIT_PET_EXPERIENCE;PLAYER_XP_UPDATE", globals = "GetPetExperience;UnitXP;UnitXPMax" }

Tags.percentxp = {[[
	if ${unit} == "pet" then
		local cur, max = GetPetExperience()
		if max > 0 then
			value = math_floor(1000 * cur/max + 0.5) / 10
		else
			value = 0
		end
	else
	 	local max = UnitXPMax(${unit})
		if max > 0 then
			local cur = UnitXP(${unit})
			value = math_floor(1000 * cur/max + 0.5) / 10
		else
			value = 0
		end
	end
]], ret = "number", events = "UNIT_PET_EXPERIENCE;PLAYER_XP_UPDATE", globals = "GetPetExperience;UnitXP;math.floor" }

Tags.missingxp = {[[
	if ${unit} == "pet" then
		local cur, max = GetPetExperience()
		value = max - cur
	else
	 	value = UnitXPMax(${unit}) - UnitXP(${unit})
	end
]], ret = "number", events = "UNIT_PET_EXPERIENCE;PLAYER_XP_UPDATE", globals = "GetPetExperience;UnitXPMax;UnitXP" }

Tags.restxp = {[[value = ${unit} == "player" and GetXPExhaustion() or 0]], ret = "number", "PLAYER_XP_UPDATE", globals = "GetXPExhaustion"}

Tags.percentrestxp = {[[do
	local max = UnitXPMax(${unit})
	if max > 0 then
		local rest = ${unit} == "player" and GetXPExhaustion() or 0
		value = math_floor(1000 * rest/max + 0.5) / 10
	else
		value = 0
	end
end]], ret = "number", events = "PLAYER_XP_UPDATE", globals = "GetXPExhaustion;UnitXPMax;math.floor" }


Tags.tappedbyme = {[[value = UnitIsTappedByPlayer(${unit}) and "*"]], ret = "string;nil", events = "Update", globals = "UnitIsTappedByPlayer" }

Tags.istapped = {[=[value = UnitIsTapped(${unit}) and not UnitIsTappedByPlayer(${unit}) and L["Tapped"]]=], ret = "string;nil", events = "Update", globals = "UnitIsTapped;UnitIsTappedByPlayer" }

Tags.ismaxhp = {[[value = UnitHealth(${unit}) == UnitHealthMax(${unit}) and 'true']], ret = "string;nil", events = "UNIT_HEALTH;UNIT_MAXHEALTH", globals = "UnitHealthMax;UnitHealth" }

Tags.ismaxmana = {[[value = UnitMana(${unit}) == UnitManaMax(${unit}) and 'true']], ret = "string;nil", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER", globals = "UnitMana;UnitManaMax" }

Tags.hasnomana = {[[value = UnitManaMax(${unit}) == 0 and 'true']], ret = "string;nil", events = "UNIT_MANA;UNIT_RAGE;UNIT_FOCUS;UNIT_ENERGY;UNIT_MAXMANA;UNIT_MAXRAGE;UNIT_MAXFOCUS;UNIT_MAXENERGY;UNIT_DISPLAYPOWER", globals = "UnitManaMax" }

Tags.incombat = {[=[value = UnitAffectingCombat(${unit}) and L['Combat']]=], ret = "string;nil", events = "Update", globals = "UnitAffectingCombat"}

Tags.fkey = {[[do
	local fkey
	if ${unit} == "player" then
		fkey = 0
	else
		fkey = (${unit}):match("^party(%d)$")
	end
	if fkey then
		value = "F" .. (fkey + 1)
	end
end]], ret = "string;nil"}

Tags.raidgroup = {[[
	for i = 1, GetNumRaidMembers() do
		local name, rank, subgroup = GetRaidRosterInfo(i)
		if name == UnitName(${unit}) then
			value = subgroup
			break
		end
	end
]], ret = "string", globals = "GetNumRaidMembers;GetRaidRosterInfo;UnitName"}

Tags.guild = {[[value = GetGuildInfo(${unit})]], ret = "string;nil", globals = "GetGuildInfo"}

Tags.guildrank = {[[do
	local _, rank = GetGuildInfo(${unit})
	value = rank
end]], ret = "string;nil", globals = "GetGuildInfo"}

Tags.target = {[[do
	local u = unit .. "target"
	if UnitExists(u) then
		value = UnitName(u)
	end
end]], ret = "string;nil", events = "UNIT_TARGET", global = "UnitExists;UnitName" }

Tags.ismouseover = {[[value = DogTag.__isMouseOver and "true"]], ret = "string;nil", events = "Mouseover" }

Tags.currep = {[[do
	local repname, repreaction, repmin, repmax, repvalue = GetWatchedFactionInfo()
	if repname then
		value = repvalue - repmin
	end
end]], ret = "number;nil", events = "Reputation", globals = "GetWatchedFactionInfo"}

Tags.maxrep = {[[do
	local repname, repreaction, repmin, repmax, repvalue = GetWatchedFactionInfo()
	if repname then
		value = repmax - repmin
	end
end]], ret = "number;nil", events = "Reputation", globals = "GetWatchedFactionInfo"}

Tags.fractionalrep = {[[do
	local repname, repreaction, repmin, repmax, repvalue = GetWatchedFactionInfo()
	if repname then
		repvalue = repvalue - repmin
		repmax = repmax - repmin
		value = repvalue .. '/' .. repmax
	end
end]], ret = "string;nil", events = "Reputation", globals = "GetWatchedFactionInfo"}

Tags.percentrep = {[[do
	local repname, repreaction, repmin, repmax, repvalue = GetWatchedFactionInfo()
	if repname then
		repmax = repmax - repmin
		repvalue = repvalue - repmin
		value = math_floor(1000 * repvalue/repmax + 0.5) / 10
	end
end]], ret = "number;nil", events = "Reputation", globals = "GetWatchedFactionInfo;math.floor"}

Tags.missingrep = {[[do
	local repname, repreaction, repmin, repmax, repvalue = GetWatchedFactionInfo()
	if repname then
		value = repmax - repvalue
	end
end]], ret = "number;nil", events = "Reputation", globals = "GetWatchedFactionInfo"}

Tags.repname = {[[value = GetWatchedFactionInfo()]], ret = "number;nil", events = "Reputation", events = "Reputation", globals = "GetWatchedFactionInfo"}

Tags.repcolor = Modifiers.repcolor
Tags.statuscolor = Modifiers.statuscolor
Tags.happycolor = Modifiers.happycolor
Tags.hostilecolor = Modifiers.hostilecolor
Tags.classcolor = Modifiers.classcolor
Tags.hpcolor = Modifiers.hpcolor
Tags.aggrocolor = Modifiers.aggrocolor
Tags.difficultycolor = Modifiers.difficultycolor
Tags.manacolor = Modifiers.manacolor
Tags.combatcolor = Modifiers.combatcolor
Tags.color = Modifiers.color
-- colors
Tags.white = Modifiers.white
Tags.red = Modifiers.red
Tags.green = Modifiers.green
Tags.blue = Modifiers.blue
Tags.cyan = Modifiers.cyan
Tags.fuchsia = Modifiers.fuchsia
Tags.yellow = Modifiers.yellow
Tags.gray = Modifiers.gray

Tags.smartmissinghp = "missinghp:hidezero:negate"

local function sortStringList(s)
	if not s then
		return nil
	end
	local list = new((";"):split(s))
	table.sort(list)
	local q = table.concat(list, ';')
	list = del(list)
	return q
end

local TagEvents = {}
local ModifierEvents = {}
local TagAliases = {}
local ModifierAliases = {}
local TagArguments = {}
local ModifierArguments = {}
local TagReturns = {}
local ModifierReturns = {}
local ModifierValues = {}
local TagGlobals = {}
local ModifierGlobals = {}
do
	for tag, data in pairs(Tags) do
		if type(data) == "table" then
			local arg = sortStringList(data.arg)
			if arg and arg ~= "nil" then
				TagArguments[tag] = arg
				local a,b,c = (";"):split(arg)
				assert(a == "nil" or a == "number" or a == "string")
				assert(not b or b == "nil" or b == "number" or b == "string")
				assert(not c or c == "nil" or c == "number" or c == "string")
			end
			TagReturns[tag] = sortStringList(data.ret)
			if data.ret then
				local a,b,c = (";"):split(data.ret)
				assert(a == "nil" or a == "number" or a == "string")
				assert(not b or b == "nil" or b == "number" or b == "string")
				assert(not c or c == "nil" or c == "number" or c == "string")
			end
			TagEvents[tag] = sortStringList(data.events)
			TagGlobals[tag] = sortStringList(data.globals)
			local func = data[1]
			if Modifiers[tag] ~= Tags[tag] then
				del(Tags[tag])
			end
			Tags[tag] = func
		elseif type(data) == "string" then
			TagAliases[tag] = data
			Tags[tag] = nil
		end
	end
	
	for mod, data in pairs(Modifiers) do
		if type(data) == "table" then
			local arg = sortStringList(data.arg)
			if arg and arg ~= "nil" then
				ModifierArguments[mod] = arg
				local a,b,c = (";"):split(arg)
				assert(a == "nil" or a == "number" or a == "string")
				assert(not b or b == "nil" or b == "number" or b == "string")
				assert(not c or c == "nil" or c == "number" or c == "string")
			end
			ModifierReturns[mod] = sortStringList(data.ret)
			if data.ret then
				local a,b,c = (";"):split(data.ret)
				assert(a == "nil" or a == "number" or a == "string" or a == "same")
				assert(not b or b == "nil" or b == "number" or b == "string" or b == "same")
				assert(not c or c == "nil" or c == "number" or c == "string" or c == "same")
			end
			local value = sortStringList(data.value)
			if value and value ~= "number;string" then
				ModifierValues[mod] = value
				local a,b = (";"):split(value)
				assert(a == "number" or a == "string")
				assert(not b or b == "number" or b == "string")
			end
			ModifierEvents[mod] = sortStringList(data.events)
			ModifierGlobals[mod] = sortStringList(data.globals)
			local func = data[1]
			del(Modifiers[mod])
			Modifiers[mod] = func
		elseif type(data) == "string" then
			ModifierAliases[mod] = data
			Modifiers[mod] = nil
		end
	end
end

local IsLegitimateUnit = {player=true,pet=true,target=true,focus=true,mouseover=true}
for i = 1, 4 do
	IsLegitimateUnit['party' .. i] = true
	IsLegitimateUnit['partypet' .. i] = true
end
for i = 1, 40 do
	IsLegitimateUnit['raid' .. i] = true
	IsLegitimateUnit['raidpet' .. i] = true
end
setmetatable(IsLegitimateUnit, {__index=function(self, unit)
	if not unit:find("target$") then
		self[unit] = false
		return false
	end
	local nonTarget = unit:sub(1, -7)
	local good = self[nonTarget]
	self[unit] = good
	return good
end})

local function toliteral(value)
	if type(value) == "number" then
		return ("%s"):format(value)
	else
		return ("%q"):format(value)
	end
end

local function unpackAndDel(t, start, finish)
	if not start then
		start = 1
	elseif not finish then
		finish = #t
	end
	if start > finish then
		t = del(t)
		return
	end
	return t[start], unpackAndDel(t, start + 1, finish)
end

local function splitCode(code)
	local alpha, bravo, charlie = code:match("^([^:!%?]*)([:!%?])(.*)$")
	if not alpha then
		return code
	end
	if alpha:find("%(") and not alpha:find("%)") then
		local t = new(splitCode(charlie))
		return alpha .. bravo .. t[1], unpackAndDel(t, 2)
	end
	return alpha, bravo, splitCode(charlie)
end

local function splitParam(code)
	local negate, tag, param = code:match("^(~*)([a-zA-Z#]+)%(([^%)]+)%)$")
	if tag then
		local t, unit = tag:match("([a-zA-Z]+)#([a-zA-Z]+)")
		if t then
			return t, param, ((negate:len() % 2) == 1 and true or false), unit:lower()
		else
			return tag, param, ((negate:len() % 2) == 1 and true or false), nil
		end
	else
		negate, tag = code:match("^(~*)(.+)$")
		if tag then
			local t, unit = tag:match("^([a-zA-Z]+)#([a-zA-Z]+)$")
			if t then
				return t, nil, ((negate:len() % 2) == 1 and true or false), unit:lower()
			else
				return tag, nil, ((negate:len() % 2) == 1 and true or false), nil
			end
		else
			return code, nil, nil, nil
		end
	end
end

local function joinParam(tag, param, negate, unit)
	local t = new()
	if negate then
		t[#t+1] = "~"
	end
	t[#t+1] = tag
	if unit then
		t[#t+1] = "#"
		t[#t+1] = unit
	end
	if param then
		t[#t+1] = "("
		t[#t+1] = param
		t[#t+1] = ")"
	end
	local s = table.concat(t)
	t = del(t)
	return s
end

local function _compileTree(code, funcs, globals)
	if type(code) == "string" then
		return code, 0
	end
	local ends
	code[1], ends = _compileTree(code[1], funcs, globals)
	if not code[1] then
		code = del(code)
		return nil, ends
	end
	local t = new()
	if not code[2] then
		local tag, param, negate, unit = splitParam(code[1])
		
		local func = Tags[tag]
		if func then
			if param then
				if not TagArguments[tag] then
					t = del(t)
					code = del(code)
					return nil, ("%q Unexpected parameter to %q"):format(param, tag)
				end
				if TagArguments[tag]:find("number") then
					param = tonumber(param) or param
					if type(param) ~= "number" and not TagArguments[tag]:find("string") then
						-- must be number
						t = del(t)
						code = del(code)
						return nil, ("%q Bad param to %q. Expected number."):format(param, tag)
					end
				end
			else
				if TagArguments[tag] and not TagArguments[tag]:find("nil") then
					t = del(t)
					code = del(code)
					return nil, ("%q Expected parameter"):format(tag)
				end
			end
			if unit and unit ~= "player" then
				if not IsLegitimateUnit[unit] then
					t = del(t)
					code = del(code)
					return nil, ("%q Unknown unit"):format(unit)
				end
				t[#t+1] = ("if UnitExists(%q) then "):format(unit)
				ends = ends + 1
			end
			local g = TagGlobals[tag]
			if g then
				g = new((";"):split(g))
				for _,v in ipairs(g) do
					globals[v] = true
				end
				g = del(g)
			end
			
			if type(func) == "function" then
				local data = new()
				data.isMod = isMod
				data.unit = unit
				data.unit_str = unit and ("%q"):format(unit) or "unit"
				data.arg = param
				data.arg_str = param and toliteral(param) or "nil"
				data.arg_string = param and ("%q"):format(param) or '"nil"'
				data.argtype = ("%q"):format(type(param))
				func = func(data)
				data = del(data)
			end
			if param then
				func = func:gsub("${arg}", toliteral(param))
				func = func:gsub("${arg:string}", ("%q"):format(param))
				func = func:gsub("${argtype}", ("%q"):format(type(param)))
			else
				func = func:gsub("${arg}", "nil")
				func = func:gsub("${arg:string}", '"nil"')
				func = func:gsub("${argtype}", '"nil"')
			end
			if unit then
				func = func:gsub("${unit}", ("%q"):format(unit))
			else
				func = func:gsub("${unit}", "unit")
			end
			t[#t+1] = func
			t[#t+1] = ";"
			if negate then
				t[#t+1] = "value = not value and 'true';"
			end
		elseif tag == "" then
			t[#t+1] = "value = "
			if negate then
				t[#t+1] = "'true'"
			else
				t[#t+1] = "nil"
			end
			t[#t+1] = ";"
		else	
			t = del(t)
			code = del(code)
			return nil, ("%q Unknown tag"):format(tag)
		end
	elseif code[2] == "?" then
		t[#t+1] = code[1]
		t[#t+1] = ("if value then value = nil;")
		local q, e = _compileTree(code[3], funcs, globals)
		if not q then
			t = del(t)
			code = del(code)
			return nil, e
		end
		t[#t+1] = q
		for i = 1, e do
			t[#t+1] = "end;"
		end
		if code[4] == "!" then
			t[#t+1] = ("else ")
			local q, e = _compileTree(code[5], funcs, globals)
			if not q then
				t = del(t)
				code = del(code)
				return nil, e
			end
			t[#t+1] = q
			for i = 1, e do
				t[#t+1] = "end;"
			end
		end
		t[#t+1] = ("end;")
	elseif code[2] == ":" then
		t[#t+1] = code[1]
		local mod, param, negate, unit = splitParam(code[3])
		local func = Modifiers[mod] or Tags[mod]
		if func then
			local isMod = func == Modifiers[mod]
			if param then
				local arg = (isMod and ModifierArguments[mod]) or (not isMod and TagArguments[mod])
				if not arg then
					t = del(t)
					code = del(code)
					return nil, ("%q Unexpected param to %q"):format(param, mod)
				end
				
				if arg:find("number") then
					param = tonumber(param) or param
					if type(param) ~= "number" and not arg:find("string") then
						-- must be number
						t = del(t)
						code = del(code)
						return nil, ("%q Bad param to %q. Expected number."):format(param, mod)
					end
				end
			else
				local arg = (isMod and ModifierArguments[mod]) or (not isMod and TagArguments[mod])
				if arg and not arg:find("nil") then
					t = del(t)
					code = del(code)
					return nil, ("%q Expected parameter"):format(mod)
				end
			end
			if unit and not IsLegitimateUnit[unit] then
				t = del(t)
				code = del(code)
				return nil, ("%q Unknown unit"):format(unit)
			end
			if isMod then
				local val = ModifierValues[mod]
				if not val then
					t[#t+1] = "if value "
					if unit and unit ~= "player" then
						t[#t+1] = "and UnitExists("
						t[#t+1] = ("%q"):format(unit)
						t[#t+1] = ") "
					end
					t[#t+1] = "then "
				else
					t[#t+1] = [[if type(value) == "]]
					t[#t+1] = val
					t[#t+1] = [[" ]]
					if unit and unit ~= "player" then
						t[#t+1] = "and UnitExists("
						t[#t+1] = ("%q"):format(unit)
						t[#t+1] = ") "
					end
					t[#t+1] = "then "
				end
			else
				t[#t+1] = "if not value "
				if unit and unit ~= "player" then
					t[#t+1] = "and UnitExists("
					t[#t+1] = ("%q"):format(unit)
					t[#t+1] = ") "
				end
				t[#t+1] = "then "
				ends = ends + 1
			end
			if negate and isMod then
				t[#t+1] = "local v = value;"
			end
			local g
			if isMod then
				g = ModifierGlobals[mod]
			else
				g = TagGlobals[mod]
			end
			if g then
				g = new((";"):split(g))
				for _,v in ipairs(g) do
					globals[v] = true
				end
				g = del(g)
			end
			if type(func) == "function" then
				local data = new()
				data.isMod = isMod
				data.unit = unit
				data.unit_str = unit and ("%q"):format(unit) or "unit"
				data.arg = param
				data.arg_str = param and toliteral(param) or "nil"
				data.arg_string = param and ("%q"):format(param) or '"nil"'
				data.argtype = ("%q"):format(type(param))
				func = func(data)
				data = del(data)
			end
			if param then
				func = func:gsub("${arg}", toliteral(param))
				func = func:gsub("${arg:string}", ("%q"):format(param))
				func = func:gsub("${argtype}", ("%q"):format(type(param)))
			else
				func = func:gsub("${arg}", "nil")
				func = func:gsub("${arg:string}", '"nil"')
				func = func:gsub("${argtype}", '"nil"')
			end
			if unit then
				func = func:gsub("${unit}", ("%q"):format(unit))
			else
				func = func:gsub("${unit}", "unit")
			end
			t[#t+1] = func
			t[#t+1] = ";"
			if negate then
				t[#t+1] = "value = not value and "
				if isMod then
					t[#t+1] = "v"
					t[#t+1] = " end;"
				else
					t[#t+1] = "'true';"
				end
			elseif isMod then
				t[#t+1] = " end;"
			end
		else
			t = del(t)
			code = del(code)
			return nil, ("%q Unknown modifier or tag"):format(mod)
		end	
	else
		t = del(t)
		code = del(code)
		return nil, ("%q Unknown separator"):format(code[2])
	end
	code = del(code)
	local s = table.concat(t)
	t = del(t)
	return s, ends
end

local function compileTree(code, funcs, globals)
	local q, e = _compileTree(code, funcs, globals)
	if not q then
		return "value = [===[[" .. e .. "]]===];"
	end
	for i = 1, e do
		q = q .. "end;"
	end
	return q
end

local function select2(min, max, ...)
	if min <= max then
		return select(min, ...), select2(min+1, max, ...)
	end
end

local function parse(...)
	local n = select('#', ...)
	if n == 1 then
		return new((...))
	end
	local q_num = 0
	local q_pos
	local x_pos
	for i = 1, n do
		if select(i, ...) == "?" then
			q_num = q_num + 1
			if not q_pos then
				q_pos = i
			end
		elseif select(i, ...) == "!" then
			q_num = q_num - 1
			if q_num == 0 then
				x_pos = i
				break
			end
		end
	end
	if q_pos then
		if not x_pos then
			return new(parse(select2(1, q_pos - 1, ...)), "?", parse(select(q_pos + 1, ...)))
		else
			return new(parse(select2(1, q_pos - 1, ...)), "?", parse(select2(q_pos + 1, x_pos - 1, ...)), "!", parse(select(x_pos + 1, ...)))
		end
	end
	return new(parse(select2(1, n - 2, ...)), ":", select(n, ...))
end

local tmp = {}
local codeToFunction = setmetatable({[""] = function() return "" end}, {__index=function(self, code)
	local t = new()
	t[#t+1] = ([[local DogTag = AceLibrary(%q);]]):format(MAJOR_VERSION)
	t[#t+1] = [[local Tags = DogTag.__Tags;local Modifiers = DogTag.__Modifiers;local tmp = DogTag.__tmp;local colors = DogTag.__colors;]]
	t[#t+1] = "" -- placeholder
	t[#t+1] = "" -- placeholder
	t[#t+1] = "local L = DogTag.__L;"
	t[#t+1] = [[return function(unit) ]]
	t[#t+1] = [[if not UnitExists(unit) then return ]]
	local globals = new()
	globals['table.concat'] = true
	globals['string.gsub'] = true
	globals['type'] = true
	globals['tonumber'] = true
	globals['UnitExists'] = true
	if code:find("%[name[%]:%?!]") then
		t[#t+1] = [[unit]]
	else
		t[#t+1] = [[""]]
	end
	t[#t+1] = [[ end;]]
	local funcs = new()
	local len = code:len()
	local st = 1
	while st <= len do
		local start, fin, left, literal, inner = code:find("^(.-)(%[([^%]]+)%])", st)
		if not start then
			local rest = code:sub(st)
			if rest:len() > 0 then
				t[#t+1] = ("tmp[#tmp+1] = %q;"):format(rest)
			end
			break
		end
		st = fin+1
		if left:len() > 0 then
			t[#t+1] = ("tmp[#tmp+1] = %q;"):format(left)
		end
		
		local u = new()
		u[#u+1] = "do local value;"
		
		local result = compileTree(parse(splitCode(inner)), funcs, globals)
		if not result then
			u = del(u)
			t[#t+1] = ("tmp[#tmp+1] = %q;"):format(literal)
		else
			u[#u+1] = result
			u[#u+1] = [[if value then tmp[#tmp+1] = value end;]]
			u[#u+1] = [[end;]]
			local s = table.concat(u)
			u = del(u)
			t[#t+1] = s
		end
	end
	local u = new()
	for k, v in pairs(funcs) do
		u[#u+1] = "local "
		u[#u+1] = k
		u[#u+1] = " = "
		u[#u+1] = v
		u[#u+1] = ";"
	end
	funcs = del(funcs)
	t[3] = table.concat(u)
	u = del(u)
	u = new()
	for k in pairs(globals) do
		u[#u+1] = "local "
		u[#u+1] = k:gsub("%.", "_")
		u[#u+1] = " = "
		u[#u+1] = k
		u[#u+1] = ";"
	end
	globals = del(globals)
	t[4] = table.concat(u)
	u = del(u)
	t[#t+1] = [[local s = table_concat(tmp);]]
	t[#t+1] = [[for i = 1, #tmp do tmp[i] = nil end;]]
	t[#t+1] = [[return (string_gsub(string_gsub(string_gsub(s, "^ +", ""), " +$", ""), "  +", " "));]]
	t[#t+1] = [[end;]]
	local s = table.concat(t)
	t = del(t)
	DogTag.__Tags = Tags
	DogTag.__Modifiers = Modifiers
	DogTag.__tmp = tmp
	DogTag.__colors = colors
	DogTag.__GetHostilityColor = GetHostilityColor
	DogTag.__GetClassColor = GetClassColor
	DogTag.__HealthGradient = HealthGradient
	DogTag.__L = L
	DogTag.__shortclasses = shortclasses
	DogTag.__shortraces = shortraces
	local val = assert(loadstring(s))()
	DogTag.__Tags = nil
	DogTag.__Modifiers = nil
	DogTag.__tmp = nil
	DogTag.__colors = nil
	DogTag.__GetHostilityColor = nil
	DogTag.__GetClassColor = nil
	DogTag.__HealthGradient = nil
	DogTag.__L = nil
	DogTag.__shortclasses = nil
	DogTag.__shortraces = nil
	self[code] = val
	return val
end, __mode='k'})

local codeToEventList = setmetatable({}, {__index = function(self, key)
	local t = new()
	local len = key:len()
	local st = 1
	while st <= len do
		local start, fin, left, literal, inner = key:find("^(.-)(%[([^%]]+)%])", st)
		if not start then
			break
		end
		st = fin+1
		local code = new(splitCode(inner))
		for i = 1, #code, 2 do
			local bit, param, negate, unit = splitParam(code[i])
			local func
			if code[i-1] ~= ":" then
				func = Tags[bit]
			else
				func = Modifiers[bit] or Tags[bit]
			end
			local isMod = func == Modifiers[bit]
			local eventList = (isMod and ModifierEvents[bit]) or (not isMod and TagEvents[bit])
			if eventList then
				local events = new((";"):split(eventList))
				for _,event in ipairs(events) do
					local e, u = event:match("([A-Z_]+)%(([a-z]+)%)")
					if e then
						t[e] = u
					else
						t[event] = unit or true
					end
				end
				events = del(events)
			end
		end
		code = del(code)
	end
	self[key] = t
	return t
end, __mode = 'k'})

local antiAlias
do
	local function handler(literal, inner)
		local code = new(splitCode(inner))
		local real = new()
		for i = 1, #code, 2 do
			local bit, param, negate, unit = splitParam(code[i])
			bit = bit:lower()
			local func
			if code[i-1] ~= ":" then
				func = TagAliases[bit]
			else
				func = ModifierAliases[bit]
				if not func and not Modifiers[bit] then
					func = TagAliases[bit]
				end
			end
			if i == 1 then
				real[1] = "["
			else
				real[#real+1] = code[i-1]
			end
			if negate then
				real[#real+1] = "~"
			end
			if func then
				if param and func:find("[:%(%?!]") then
					if func:find("%(arg%)") then
						real[#real+1] = func:gsub("%(arg%)", "(" .. param .. ")")
					else
						real[#real+1] = joinParam(bit, param, nil, unit)
					end
				elseif unit and func:find("#") then
					real[#real+1] = joinParam(bit, param, nil, unit)
				elseif not func:find("[:%?!]") then
					local fbit, fparam, fnegate, funit = splitParam(func)
					real[#real+1] = joinParam(fbit, fparam or param, fnegate, funit or unit)
				elseif unit then
					if param then
						real[#real+1] = joinParam(bit, param, nil, unit)
					else
						local fcode = new(splitCode(func))
						for j = 1, #fcode, 2 do
							if j > 1 then
								real[#real+1] = fcode[j-1]
							end
							local fbit, fparam, fnegate, funit = splitParam(fcode[j])
							real[#real+1] = joinParam(fbit, fparam, fnegate, unit)
						end
					end
				else
					real[#real+1] = func
					if param then
						real[#real+1] = "("
						real[#real+1] = param
						real[#real+1] = ")"
					end
				end
			else	
				real[#real+1] = joinParam(bit, param, nil, unit)
			end
		end
		code = del(code)
		
		real[#real+1] = ']'
		local str = table.concat(real):gsub("~~", "")
		real = del(real)
		return str
	end
	antiAlias = setmetatable({}, {__index = function(self, key)
		local value = key
		while true do
			local lastValue = value
			value = lastValue:gsub("(%[([^%]]+)%])", handler)
			if value == lastValue then
				break
			end
		end
		self[key] = value
		return value
	end, __mode= 'k'})
end

local eventData = setmetatable({}, {__index = function(self, event)
	local t = new()
	self[event] = t
	if event ~= "Update" and event ~= "Mouseover" and event ~= "Reputation" then
		DogTag.frame:RegisterEvent(event)
	end
	return t
end})
local buckets = setmetatable({}, {__index = function(self, event)
	local t = new()
	self[event] = t
	return t
end})

local registry
local codeHash

local frameToText = {}
local textToFrame = {}

local NormalUnits = {player = true, pet = true, target = true, mouseover = true, focus = true}
for i = 1, 4 do
	NormalUnits['party' .. i] = true
	NormalUnits['partypet' .. i] = true
end
for i = 1, 40 do
	NormalUnits['raid' .. i] = true
end
local deprecated = tonumber(date("%Y%m%d")) > 20070518
function DogTag:RegisterFontString(text, frame, unit, code)
	if not code and not deprecated then
		unit, code = frame, unit
		self:argCheck(text, 2, "table")
		self:argCheck(unit, 3, "string")
		self:argCheck(code, 4, "string")
	else
		self:argCheck(text, 2, "table")
		self:argCheck(frame, 3, "table")
		self:argCheck(unit, 4, "string")
		self:argCheck(code, 5, "string")
	end
	if not IsLegitimateUnit[unit] then
		code = ""
	end
	code = antiAlias[code]
	if codeHash[text] then
		if codeHash[text] == code and registry[text] == unit then
			if text:IsShown() and textToFrame[text]:IsShown() then
				DogTag.__isMouseOver = frame == lastMouseFocus
				text:SetText(codeToFunction[code](unit))
			end
			return
		end
		self:UnregisterFontString(text)
	end
	codeHash[text] = code
	registry[text] = unit
	if NormalUnits[unit] then
		local events = codeToEventList[code]
	
		for event, u in pairs(events) do
			if u == true then
				eventData[event][text] = unit
			else
				eventData[event][text] = u
			end
		end
	end
	registry[text] = unit
	
	if frame then
		if not frameToText[frame] then
			frameToText[frame] = new()
		end
		frameToText[frame][text] = unit
		textToFrame[text] = frame
	else
		textToFrame[text] = text
	end
	
	if text:IsShown() and textToFrame[text]:IsShown() then
		DogTag.__isMouseOver = frame == lastMouseFocus
		text:SetText(codeToFunction[code](unit))
	end
end

function DogTag:UnregisterFontString(text)
	self:argCheck(text, 2, "table")
	local code = codeHash[text]
	if not code then
		return
	end
	
	local unit = registry[text]
	registry[text] = nil
	codeHash[text] = nil
	if NormalUnits[unit] then
		local events = codeToEventList[code]
	
		for event in pairs(events) do
			local t = eventData[event]
			assert(t)
			t[text] = nil
			if not next(t) then
				eventData[event] = del(t)
				if event ~= "Update" and event ~= "Mouseover" and event ~= "Reputation" then
					self.frame:UnregisterEvent(event)
					buckets[event] = del(buckets[event])
				end
			end
		end
	end
	
	text:SetText("")
	
	local frame = textToFrame[text]
	if frame == text then
		textToFrame[text] = nil
	elseif frame then
		textToFrame[text] = nil
		frameToText[frame][text] = nil
		if not next(frameToText[frame]) then
			frameToText[frame] = del(frameToText[frame])
		end
	end
end

function DogTag:UpdateFontString(text)
	self:argCheck(text, 2, "table")
	local unit = registry[text]
	local frame = textToFrame[text]
	if unit and text:IsShown() and frame:IsShown() then
		DogTag.__isMouseOver = frame == lastMouseFocus
		text:SetText(codeToFunction[codeHash[text]](unit))
	end
end

function DogTag:UpdateAllForUnit(unit)
	self:argCheck(unit, 2, "string")
	if not IsLegitimateUnit[unit] then
		return
	end
	
	local toUpdate = new()
	for text, u in pairs(registry) do
		if u == unit then
			toUpdate[text] = true
		end
	end
	for event, data in pairs(eventData) do
		for text, u in pairs(data) do
			if u == unit then
				toUpdate[text] = true
			end
		end
	end
	for text in pairs(toUpdate) do
		local frame = textToFrame[text]
		if text:IsShown() and frame:IsShown() then
			DogTag.__isMouseOver = frame == lastMouseFocus
			text:SetText(codeToFunction[codeHash[text]](unit))
		end
	end
	toUpdate = del(toUpdate)
end

function DogTag:UpdateAllForFrame(frame)
	self:argCheck(frame, 2, "table")
	
	local list = frameToText[frame]
	if not list then
		return
	end
	
	for text, unit in pairs(list) do
		local frame = textToFrame[text]
		if text:IsShown() and frame:IsShown() then
			DogTag.__isMouseOver = frame == lastMouseFocus
			text:SetText(codeToFunction[codeHash[text]](unit))
		end
	end
end

function DogTag:Evaluate(unit, code)
	if not IsLegitimateUnit[unit] then
		return ""
	end
	DogTag.__isMouseOver = false
	return codeToFunction[antiAlias[code]](unit)
end

function DogTag:PARTY_MEMBERS_CHANGED()
	for text, unit in pairs(registry) do
		local frame = textToFrame[text]
		if text:IsShown() and frame:IsShown() then
			DogTag.__isMouseOver = frame == lastMouseFocus
			text:SetText(codeToFunction[codeHash[text]](unit))
		end
	end
end
DogTag.PARTY_LEADER_CHANGED = DogTag.PARTY_MEMBERS_CHANGED
DogTag.RAID_ROSTER_UPDATE = DogTag.PARTY_MEMBERS_CHANGED
DogTag.PARTY_MEMBER_ENABLE = DogTag.PARTY_MEMBERS_CHANGED
DogTag.PARTY_MEMBER_DISABLE = DogTag.PARTY_MEMBERS_CHANGED
DogTag.PARTY_ENTERING_WORLD = DogTag.PARTY_MEMBERS_CHANGED
DogTag.UNIT_PET = DogTag.PARTY_MEMBERS_CHANGED

function DogTag:PLAYER_TARGET_CHANGED()
	self:UpdateAllForUnit('target')
end

function DogTag:PLAYER_FOCUS_CHANGED()
	self:UpdateAllForUnit('focus')
end

function DogTag:PLAYER_PET_CHANGED()
	self:UpdateAllForUnit('pet')
end

function DogTag:UPDATE_MOUSEOVER_UNIT()
	self:UpdateAllForUnit('mouseover')
end

function DogTag:PLAYER_LOGIN()
	-- for extra addons' help, such as MobHealth3 or MobInfo2
	for k in pairs(codeToFunction) do
		codeToFunction[k] = nil
	end
	self:PARTY_MEMBERS_CHANGED()
end

function DogTag:SetColorConstantTable(t)
	colors = t
	-- need to update the colors reference in the functions
	for k in pairs(codeToFunction) do
		codeToFunction[k] = nil
	end
	self:PARTY_MEMBERS_CHANGED()
end

function DogTag:OnMouseoverUpdate()
	local mouseFocus = GetMouseFocus()
	local list = frameToText[lastMouseFocus]
	lastMouseFocus = mouseFocus
	if list then
		DogTag.__isMouseOver = false
		for text, unit in pairs(eventData.Mouseover) do
			if list[text] and text:IsShown() and textToFrame[text]:IsShown() then
				text:SetText(codeToFunction[codeHash[text]](unit))
			end
		end
	end
	list = frameToText[mouseFocus]
	if list then
		DogTag.__isMouseOver = true
		for text, unit in pairs(eventData.Mouseover) do
			if list[text] and text:IsShown() and textToFrame[text]:IsShown() then
				text:SetText(codeToFunction[codeHash[text]](unit))
			end
		end
	end
end

function DogTag:OnReputationUpdate()
	for text, unit in pairs(eventData.Reputation) do
		if text:IsShown() and textToFrame[text]:IsShown() then
			text:SetText(codeToFunction[codeHash[text]](unit))
		end
	end
end

local function activate(self, oldLib, oldDeactivate)
	DogTag = self
	
	self.frame = oldLib and oldLib.frame or CreateFrame("Frame", "DogTag10Frame")
	
	self.frame:UnregisterAllEvents()
	self.frame:RegisterEvent("PARTY_MEMBERS_CHANGED")
	self.frame:RegisterEvent("PARTY_LEADER_CHANGED")
	self.frame:RegisterEvent("RAID_ROSTER_UPDATE")
	self.frame:RegisterEvent("PARTY_MEMBER_ENABLE")
	self.frame:RegisterEvent("PARTY_MEMBER_DISABLE")
	self.frame:RegisterEvent("PARTY_ENTERING_WORLD")
	self.frame:RegisterEvent("PLAYER_TARGET_CHANGED")
	self.frame:RegisterEvent("PLAYER_FOCUS_CHANGED")
	self.frame:RegisterEvent("PLAYER_PET_CHANGED")
	self.frame:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
	self.frame:RegisterEvent("UNIT_PET")
	self.frame:RegisterEvent("PLAYER_LOGIN")
	self.frame:SetScript("OnEvent", function(this, event, arg1)
		local func = self[event]
		if func then
			func(self)
			return
		end	
		buckets[event][arg1 or false] = true
	end)
	local timePassed = 0
	local num = 0
	local toUpdate = {}
	self.frame:SetScript("OnUpdate", function(this, elapsed)
		timePassed = timePassed + elapsed
		if timePassed > 0.05 then
			timePassed = 0
			num = (num + 1) % 3
			if num == 0 then
				if eventData.Update then
					for text in pairs(eventData.Update) do
						toUpdate[text] = true
					end
				end
			end
			for event, units in pairs(buckets) do
				if units[false] then
					for text in pairs(eventData[event]) do
						toUpdate[text] = true
					end
				else
					for text, unit in pairs(eventData[event]) do
						if units[unit] then
							toUpdate[text] = true
						end
					end
				end
				buckets[event] = del(units)
			end
			for text in pairs(toUpdate) do
				toUpdate[text] = nil
				local unit = registry[text]
				local frame = textToFrame[text]
				if unit and text:IsShown() and frame:IsShown() then
					DogTag.__isMouseOver = frame == lastMouseFocus
					text:SetText(codeToFunction[codeHash[text]](unit))
				end
			end
		end
	end)
	
	if not oldLib or not oldLib.hookedRep then
		hooksecurefunc("ReputationWatchBar_Update", function()
			self:OnReputationUpdate()
		end)
	end
	self.hookedRep = true
	
	registry = {}
	self.registry = registry
	codeHash = {}
	self.codeHash = codeHash
	if oldLib then
		for text, unit in pairs(oldLib.registry) do
			self:RegisterFontString(text, unit, oldLib.codeHash[text])
		end
	end
end

AceLibrary:Register(DogTag, MAJOR_VERSION, MINOR_VERSION, activate, nil, external)
